//
// Institute for Signal Processing (University of Luebeck, Germany)
// Copyright (c) 2010 by Florian Mueller
//
// Permission to use, copy, modify, and distribute this software without
// fee is hereby granted FOR RESEARCH/EDUCATION PURPOSES only, provided
// that this copyright notice appears in all copies and in all supporting
// documentation, and that the software is not redistributed for any
// fee (except for a nominal shipping charge).
//
// For any other uses of this software, in original or modified form,
// including but not limited to consulting, production or distribution
// in whole or in part, specific prior permission must be obtained
// from the author.
// Signal processing methods and algorithms implemented by this
// software may be claimed by patents owned by others.
//
// The author makes no representation about the suitability of this
// software for any purpose. It is provided "as is" without warranty
// of any kind, either expressed or implied.
// Beware of the bugs.
//
//     Revision history
//
//     Ver     Date         Description
//     -------------------------------------------------------------------
//     0.9     01-05-2010   basic version 
//     0.9.2   01-07-2010   Recoded in C++ with OOP approach
//

#include "Filelist.h"
/***************
 * Constructor *
 ***************/
Filelist::Filelist()
{
    currentIndex=0;
    filename="";
    LittleEndian=true;
}

Filelist::~Filelist()
{
    sData.clear();
}

Filelist::Filelist(string listenfilename)
{
    isBigEndian();
    currentIndex=0;
    filename=listenfilename;
    string tempString;
    ifstream file;
    file.open(filename.c_str(), ios::in);
    if (file.good())
    {
        int i=0;
        while ((getline(file, tempString))&&(!tempString.empty()))
        {
            sData.push_back(tempString);
            i++;
        }
    }
    else
    {
        printf("ERROR: Cannot open %s for reading. \n",listenfilename.c_str());
        file.close();
        exit(10);
    }

    file.close();

}


/**************************************************
 *                  isBigEndian                   *
 * check endianness of system                     *
 *                                                *
 **************************************************/


void Filelist::isBigEndian()
{
    int i = 1;
    char *p = (char *) &i;
    if (p[0] == 1)
    {
        LittleEndian=true;
    }
    else
    {
        LittleEndian=false;
    }
}
/**************************************************
 *                  getSignal                     *
 * read wave file to double vector                *
 *                                                *
 * OUTPUT:               dVec  result             *
 *                                                *
 **************************************************/


dVec Filelist::getSignal()
{
    return getSignal(currentIndex);
}
/**************************************************
 *                  getSignal                     *
 * read wave file to double vector                *
 *                                                *
 * INPUT:               int  position             *
 * OUTPUT:              dVec   result             *
 *                                                *
 **************************************************/
dVec Filelist::getSignal(int index)
{
    dVec result;
    short val;
    stringstream ss;
    //open file
    fstream fin(sData[index].c_str(), ios::binary| ios::in);
    //read signal values
    if (fin.good())
    {
        while (fin.read((char*)&val, 2))
        {
            result.push_back(endian_swap(val));  // copy to vector
        }
    }
    else
    {
        printf("Can not open file %s for reading\n", sData[index].c_str());
        fin.close();

        exit(11);
    }
    fin.close();
    return result;
}

int Filelist::length()
{
    return sData.size();
}

string Filelist::getIterator()
{
    if (currentIndex>=sData.size())
    {
        exit(100);
    }
    return sData[currentIndex];
}

void Filelist::getNext()
{
    currentIndex++;

}


/************************************************************
 *                       readhtk                            *
 * read file (filename). Store htk-values in matrix (data)  *
 * If necessary with endian swap.                           *
 * Note: htk-header: htk-type will not stored               *
 *                                                          *
 * INPUT:       string                      htkfile         *
 * IN/OUTPUT:   & double                    frameshift      *
 *              & vector<vector<double>>    data            *
 *                                                          *
 ************************************************************/
Array2d<double> Filelist::readhtk(double frameshift)
{
    return readhtk(sData[currentIndex], frameshift);
}


Array2d<double> Filelist::readhtk(string filename, double frameshift)
{
    float* temp;
    temp = new float[1];
    unsigned int i, j, nSamples, sPeriod;
    unsigned short sSize, sKind;
    FILE *fp;
    unsigned long int length;
    if ((fp=fopen(filename.c_str(), "rb"))==NULL) // Open htk-file
    {
        printf("Cannot open file %s for reading\n", filename.c_str());
        exit(40);
    }
    /***************
     * read header *
     ***************/
    if (fread(&nSamples, sizeof(nSamples),1, fp) != 1)
    {
        fclose(fp);
        printf("Error reading %s \n (Header problem)\n", filename.c_str());
        exit(41);
    }
    nSamples=endian_swap(nSamples);//read number of frames, swap endian if necessary
    if (fread(&sPeriod, sizeof(sPeriod),1, fp) != 1)
    {
        fclose(fp);
        printf("Error reading %s \n (Header problem)\n", filename.c_str());
        exit(41);
    }
    frameshift=endian_swap(sPeriod)*1.0000e-07;//read frame interval (converted to ms), swap endian if necessary
    if (fread(&sSize, sizeof(sSize),1, fp) != 1)
    {
        fclose(fp);
        printf("Error reading %s \n (Header problem)\n", filename.c_str());
        exit(41);
    }
    sSize=endian_swap(sSize);//read bytes per frame, swap endian if necessary
    if (fread(&sKind, sizeof(sKind),1, fp) != 1)
    {
        fclose(fp);
        printf("Error reading %s \n (Header problem)\n", filename.c_str());
        exit(41);
    }
    sKind=endian_swap(sKind);//read htk typ, this value will get lost, write htk in type 9: user defined features
    length=nSamples*sSize/4;
    /***************
     * read values *
     ***************/
    Array2d<double> data(nSamples,sSize/4);

    for (i=0; i<nSamples;i++)
    {
        for (j=0; j<sSize/4; j++)
        {
            if (fread(temp, sizeof(float), 1, fp)!=1)
            {
                printf("Error during reading htk-file %s\n", filename.c_str());
                exit(42);
            }
            data[i][j]=(endian_swap(*temp)); //Read value and endian swap if necessary
        }
    }
    fclose(fp);
    delete[] temp;
    return data;
}


/********************************************************
 *                      writehtk                        *
 * write htk header and values (data) to file (htkfile) *
 * If necessary with endian swap.                       *
 * Note:    htk-type will always set to 9: user-format  *
 *          header informations from parameter          *
 *          (frameshift) and matrix dimension (data)    *
 *                                                      *
 * INPUT:       string                      htkfile     *
 *              double                      frameshift  *
 * IN/OUTPUT:   & vector<vector<double>>    data        *
 *                                                      *
 ********************************************************/
void Filelist::writehtk(Array2d<double> data, double frameshift)
{
    writehtk(sData[currentIndex], data, frameshift);
}


void Filelist::writehtk(std::string filename, Array2d<double> data, double frameshift)
{
    FILE *fp;
    float temp;
    unsigned int i, j, tempInt;
    unsigned short tempShort;
    /* Open file for write. */
    if ((fp = fopen(filename.c_str(),"wb")) == NULL)
    {
        printf("ERROR: Cannot open file %s for writing.\n\tMake sure the directory exists and you have the security privileges for the whole path\n", filename.c_str());
        fclose(fp);
        exit(50);
    }
    /****************
     * write header *
     ****************/
    tempInt=data.getZeilen();
    tempInt=endian_swap(tempInt);
    if (fwrite(&tempInt, sizeof(tempInt), 1, fp) != 1) // write number of frames
    {
        printf("ERROR: Cannot write HTK header to file %s.\n", filename.c_str());
        fclose(fp);
        exit(51);
    }
    tempInt=(unsigned int) floor((frameshift*10000000)*0.5);
    tempInt=endian_swap(tempInt);
    if (fwrite(&tempInt, sizeof(tempInt), 1, fp) != 1) //write frame interval (converted to seconds)
    {
        printf("ERROR: Cannot write HTK header to file %s.\n", filename.c_str());
        fclose(fp);
        exit(51);
    }
    tempShort=data.getSpalten()*4;
    tempShort=endian_swap(tempShort);
    if (fwrite(&tempShort, sizeof(tempShort), 1, fp) != 1)  // write bytes per frame
    {
        printf("ERROR: Cannot write HTK header to file %s.\n", filename.c_str());
        fclose(fp);
        exit(51);
    }
    tempShort=9;
    tempShort=endian_swap(tempShort);
    if (fwrite(&tempShort, sizeof(tempShort), 1, fp) != 1)
    {
        printf("ERROR: Cannot write HTK header to file %s.\n", filename.c_str()); // write type code 9: user defined features
        fclose(fp);
        exit(51);
    }
    /****************
     * write values *
     ****************/
    for (i=0; i<data.getZeilen();i++)     //2
    {
        for (j=0; j<data.getSpalten(); j++) //395
        {
            temp=(float)data[i][j];
            temp=endian_swap(temp);

            if (fwrite(&temp, sizeof(float), 1, fp)!= 1)
            {
                printf("ERROR: Cannot write HTK data to file %s.\n", filename.c_str());
                fclose(fp);
                exit(51);
            }
        }
    }
    fclose(fp);
}


/**********************************************************************************************
 *                                  Endian swap methodes                                      *
 **********************************************************************************************/
float Filelist::endian_swap(float d)
{
    if (LittleEndian)
    {
        float         a;
        unsigned char *dst = (unsigned char *)&a;
        unsigned char *src = (unsigned char *)&d;
        dst[0] = src[3];
        dst[1] = src[2];
        dst[2] = src[1];
        dst[3] = src[0];
        return a;
    }
    else
    {
        return d;
    }
}

double Filelist::endian_swap(double d)
{
    if (LittleEndian)
    {
        double         a;
        unsigned char *dst = (unsigned char *)&a;
        unsigned char *src = (unsigned char *)&d;
        dst[0] = src[7];
        dst[1] = src[6];
        dst[2] = src[5];
        dst[3] = src[4];
        dst[4] = src[3];
        dst[5] = src[2];
        dst[6] = src[1];
        dst[7] = src[0];
        return a;
    }
    else
    {
        return d;
    }
}


unsigned int Filelist::endian_swap(unsigned int& i)
{
    if (LittleEndian)
    {
        i = (i>>24) |
            ((i<<8) & 0x00FF0000) |
            ((i>>8) & 0x0000FF00) |
            (i<<24);
    }
    return i;
}

short Filelist::endian_swap( short s )
{
    if (LittleEndian)
    {
        unsigned char b1, b2;

        b1 = s & 255;
        b2 = (s >> 8) & 255;
        return (b1 << 8) + b2;
    }
    return s;
}

unsigned short Filelist::endian_swap(unsigned short& x)
{
    if (LittleEndian)
    {
        x = (x>>8) |
            (x<<8);
    }
    return x;
}

/*
unsigned long int Filelist::endian_swap(unsigned long int& x)
{
    if (LittleEndian)
    {
        x = (x>>56) |
            ((x<<40) & 0x00FF000000000000) |
            ((x<<24) & 0x0000FF0000000000) |
            ((x<<8)  & 0x000000FF00000000) |
            ((x>>8)  & 0x00000000FF000000) |
            ((x>>24) & 0x0000000000FF0000) |
            ((x>>40) & 0x000000000000FF00) |
            (x<<56);
    }
    return x;
}

*/


