//
// Institute for Signal Processing (University of Luebeck, Germany)
// Copyright (c) 2010 by Florian Mueller
//
// Permission to use, copy, modify, and distribute this software without
// fee is hereby granted FOR RESEARCH/EDUCATION PURPOSES only, provided
// that this copyright notice appears in all copies and in all supporting
// documentation, and that the software is not redistributed for any
// fee (except for a nominal shipping charge).
//
// For any other uses of this software, in original or modified form,
// including but not limited to consulting, production or distribution
// in whole or in part, specific prior permission must be obtained
// from the author.
// Signal processing methods and algorithms implemented by this
// software may be claimed by patents owned by others.
//
// The author makes no representation about the suitability of this
// software for any purpose. It is provided "as is" without warranty
// of any kind, either expressed or implied.
// Beware of the bugs.
//
//     Revision history
//
//     Ver     Date         Description
//     -------------------------------------------------------------------
//     0.9     01-05-2010   basic version 
//     0.9.2   01-07-2010   Recoded in C++ with OOP approach
//

#include "GammatoneFB.h"
using namespace std;

GammatoneFB::GammatoneFB()
{
    //ctor
}

GammatoneFB::~GammatoneFB()
{
    //dtor
}


Array2d<double> GammatoneFB::computeGTFOutput(dVec signal, double samplerate, double framesize, double frameshift, double width, double preemphasizeCoef, double nfilter, double freqmin, double freqmax, double compression, bool meannorm, bool stdnorm, string CFfile)
{
    dVec freqcenters;
    Array2d<double> frames;
    signal=rmsNorm(signal);
    frames= makeFrames(frameshift ,framesize, samplerate ,signal);
    frames = removeDV(frames);
    frames=hamming(frames);// use hamming window filter
    if (preemphasizeCoef != 1)
    {
        frames=preemphasize(preemphasizeCoef,frames);
    }
    frames = doFFT(frames);
    if (!CFfile.empty())                        // get CenterFrequenz from file
    {
        freqcenters = readCF(CFfile);
        if (freqcenters.size()>nfilter)
        {
            cout<<"WARNING: dimension mismatch: "<<freqcenters.size()<<" center frequencys but only "<<nfilter<<" filters"<<endl<<"            use the first "<<nfilter<<" values of "<<CFfile<<endl;
            freqcenters.resize(nfilter);
        }
        else if (freqcenters.size()<nfilter)
        {
            cout<<"ERROR: dimension mismatch: given "<<freqcenters.size()<<" center frequencys but "<<nfilter<<" needed"<<endl<<"            change number of filters or center frequencys file"<<endl;
            exit(13);
        }
    }
    else                                        // OR  generatate center frequenzes
    {
        freqcenters = calcCenterFrequencies(nfilter, freqmin, freqmax, freqcenters);
    }

    Array2d<double> weights = calcWeights(2*frames.getSpalten(), nfilter, freqmin, freqmax, samplerate, width, freqcenters, meannorm, stdnorm);

    Array2d<double> gammatoneMat = sMult(weights,frames);
    gammatoneMat = transAndFac(gammatoneMat, compression);
    freqcenters.clear();
    return gammatoneMat;

}

Array2d<double>  GammatoneFB::transAndFac(Array2d<double> mat, double compression)
{
    Array2d<double>  mat2(mat.getSpalten(),mat.getZeilen());
    unsigned int i;
    unsigned int j;

    for (i=0;i<mat2.getZeilen();i++)
    {
        for (j=0;j<mat.getZeilen();j++)
        {
            mat2[i][j]=pow(mat[j][i],compression);
        }
    }
    return mat2;
}


/***********************************************************
 *                        sMult                            *
 * special Matrix multiplication : (1/size_A)*A*B=C        *
 *                                                         *
 * INPUT:     Array2d<double>          A                   *
 *            Array2d<double>          B                   *
 * OUTPUT:    Array2d<double>          C                   *
 *                                                         *
 * June 2010  Peter-Christian Quint, pcq@gmx.de            *
 ***********************************************************/
Array2d<double> GammatoneFB::sMult(Array2d<double>  A, Array2d<double>  B)
{

    Array2d<double> C (A.getZeilen(),B.getZeilen());
    double a=2*(A.getSpalten()-1);
    a=1/a;
    unsigned int i, j, k;
    double temp;
    for (i=0; i<A.getZeilen(); i++)
    {
        for (j=0; j<B.getZeilen(); j++)
        {
            temp=0;
            for (k=0; k<A.getSpalten()-1; k++)
            {
                temp+=A[i][k+1]*B[j][k];
            }
            C[i][j]=temp*a;
        }
    }
    return C;
}

/************************************************************
 *                  calcCenterFrequencies                   *
 * calculate center frequencey with Slaney's MakeERBFilters *
 * nfilts: number of used filters                           *
 * minfreq: minimum frequeny                                *
 * maxfreq: maximum frequeny                                *
 * cfreqs:  output vector                                   *
 *                                                          *
 * INPUT:       double                      nfilts          *
 *              double                      minfreq         *
 *              double                      maxfreq         *
 * OUTPUT:      vector<double>              cfreqs          *
 *                                                          *
 * June 2010  Peter-Christian Quint, pcq@gmx.de             *
 ************************************************************/
dVec GammatoneFB::calcCenterFrequencies (double nfilts, double minfreq, double maxfreq, dVec cfreqs)
{
    double EarQ=9.26449;
    double minBW = 24.7;
    double i;
    double nf = nfilts-1;
    double temp1 = -1*EarQ*minBW;
    double temp3 = maxfreq + EarQ*minBW;
    double temp2 = (-1*log(temp3)+log(minfreq + EarQ*minBW))/nf;
    cfreqs.reserve(nfilts);
    if (cfreqs.empty())         //if not set by paramters: calculate nfilts center frequencies within the freqeuency range [minfreq,maxfreq]
    {
        for (i=0;i<nfilts; i++)
        {
            cfreqs.push_back(temp1+exp((nf-i)*temp2)*temp3);
        }
    }
    return cfreqs;
}

/**************************************************
 *                 readCF                         *
 * read center frequenzes to double vector        *
 *                                                *
 * INPUT:  (*char)             filename           *
 * OUTPUT: (vector<double>)    result             *
 *                                                *
 * Dec 2009  Peter-Christian Quint, pcq@gmx.de    *
 **************************************************/
dVec GammatoneFB::readCF(string filename)
{
    dVec result;
    short val;
    stringstream ss;
    //open file
    fstream fin(filename.c_str(), ios::binary| ios::in);
    //read signal values
    if (fin.good())
    {
        while (fin.read((char*)&val, 2))
        {
            result.push_back(val);  // copy to vector
        }
    }
    else
    {
        printf("Can not open file %s for reading\n", filename.c_str());
        fin.close();

        exit(11);
    }
    fin.close();
    return result;
}
/****************************************************
 *                  preemphasize                    *
 * filter signal (frames) with 	coefficient (coef)  *
 * with input signal (frames)                       *
 *                                                  *
 * INPUT:       Array2d<double>             frames  *
 *              double                      coef    *
 *                                                  *
 * OUTPUT:      Array2d<double>             frames  *
 *                                                  *
* Dec 2009  Peter-Christian Quint, pcq@gmx.de       *
 ****************************************************/
Array2d<double> GammatoneFB::preemphasize(double coef, Array2d<double> frames)
{
    unsigned int i;
    unsigned int j;
    for (i=0;i<frames.getZeilen();i++)
    {
        for (j=1; j<frames.getSpalten();j++)
        {
            frames[i][j]=frames[i][j]+coef*frames[i][j-1];
        }
        frames[i][0]=(1+coef)*frames[i][1];
    }
    return frames;
}

/************************************************
 *                   rmsNorm                    *
 * calculate root mean square and normalizing   *
 * data in codomain of 16bit signed integer     *
 *                                              *
 * IN/OUTPUT:  vector<double>     data          *
 *                                              *
 * Dec 2009  Peter-Christian Quint, pcq@gmx.de  *
 ************************************************/
dVec GammatoneFB::rmsNorm(dVec data)
{
    double max=0;
    unsigned int i;
    double sum=0;
    for (i=0; i<data.size();i++)
    {
        sum+=pow(data[i],2);
    }
    for (i=0; i<data.size();i++)
    {
        data[i]=data[i]/sqrt(sum/data.size());
        if (abs(data[i])>max)
        {
            max=abs(data[i]);
        }
    }
    for (i=0; i<data.size();i++)
    {
        data[i]=data[i]/max*32768;
    }
    return data;
}

/************************************************
 *               Hamming window                 *
 * calculate hamming window to multiply it      *
 * with input signal (frames)                   *
 *                                              *
 * IN/OUTPUT: Array2d<double>           frames  *
 *                                              *
 * Dec 2009  Peter-Christian Quint, pcq@gmx.de  *
 ************************************************/
Array2d<double> GammatoneFB::hamming(Array2d<double> frames)
{
    double  pi = 3.1415926535897932384626433832795028841971693993751058209749445923078164062862089986280348253421170679; //approx of PI
    double N=(double)frames.getSpalten();
    int n;
    unsigned int j;
    double hamming;
    for (n=0;n<N;n++)
    {
        hamming=0.54-0.46*cos((2*pi*n)/(N-1));
        for (j=0;j<frames.getZeilen();j++)
        {
            frames[j][n]*=hamming;
        }
    }

    return frames;
}

/***********************************************************
 *                      makeFrames                         *
 * Convert signal to frames                                *
 *                                                         *
 * INPUT:     double                    frameshift         *
 *            double                    framesize          *
 *            double                    samplerate         *
 *            vector<double>            data               *
 * OUTPUT:    Array2d<double>         frames               *
 *                                                         *
 * June 2010  Peter-Christian Quint, pcq@gmx.de            *
 ***********************************************************/

Array2d<double> GammatoneFB::makeFrames(double frameshift,double framesize,double samplerate,dVec signal)
{
    unsigned int j;
    int i,flow=0;
    unsigned int dimFrame=(unsigned int) (samplerate*framesize);                                   // size of Frames in numbers
    unsigned int dimFrameShift=(unsigned int) (samplerate*frameshift);                             // size of Frameshift in numbers
    if (dimFrame<dimFrameShift)
    {
        cout<<"ERROR: Frameshift can not be bigger than Framesize"<<endl;
        exit(20);
    }

    int numberFrames= floor(signal.size()/(dimFrame-dimFrameShift))-1;

    Array2d<double> frames(numberFrames, dimFrame);
    for (i=0;i<numberFrames;i++)
    {
        for (j=0;j<dimFrame;j++)
        {
            frames[i][j]=signal[j+flow];
        }
        flow+=dimFrame-dimFrameShift;
    }
    return frames;
}

/*************
 * remove DV *
 *************/

Array2d<double> GammatoneFB::removeDV(Array2d<double> frames)
{
    double mean=0;
    unsigned int i,j;
    for (i=0;i<frames.getZeilen();i++)
    {

        for (j=0;j<frames.getSpalten();j++)
        {
            mean+=frames[i][j];
        }
        mean/=frames.getSpalten();
        for (j=0;j<frames.getSpalten();j++)
        {
            frames[i][j]-=mean;
        }
        mean=0;
    }
    return frames;
}

/************************************************************************************
 *                              calcWeights                                         *
 * special Matrix multiplication : (1/size_A)*A*B=C                                 *
 *                                                                                  *
 * INPUT:     int                       nfft             number ffts                *
 *            double                    nfilters         number of filters          *
 *            double                    freqmin          minimum frequency          *
 *            double                    freqmax          maximum frequency          *
 *            double                    samplerate       samplerate                 *
 *            double                    width            frame width                *
 *            vector<double>            freqcenters      frquenzy centers           *
 *            bool                      meannorm         true for norm with mean    *
 *            bool                      stdnorm          true for standardnorm      *
 * OUTPUT     Array2d<double>           wts              weighs matrix              *
 *                                                                                  *
 *                                                                                  *
 * Generate a matrix of weights to combine FFT bins into Gammatone bins.            *
 * nfft defines the source FFT size at samplerate. nfilts specifies the             *
 * number of output bands required, and width is the constant width of each         *
 * band in Bark. minfreq, maxfreq specify range covered in Hz. While wts            *
 * has nfft columns, the second half are all zero.                                  *
 *                                                                                  *
 *** matlab template by                                                             *
 * 2004-09  Dan Ellis dpwe@ee.columbia.edu  based on rastamat/audspec.m             *
 * 2009-06  matlab file modificated by Florian Mueller mueller@isip.uni-luebeck.de  *
 ***                                                                                *
 * Feb. 2010  matlab code to C++ code by Peter-Christian Quint, pcq@gmx.de          *
 *                                                                                  *
 ************************************************************************************/
Array2d<double> GammatoneFB::calcWeights(int nfft, double nfilter, double freqmin, double freqmax, double samplerate, double width, dVec freqcenters, bool meannorm, bool stdnorm)
{
    Array2d<double> wts (freqcenters.size(),(nfft/2)+1);
    cVec ucirc, zros;
    double EarQ=9.26449;
    double minBW= 24.7;
    double pi=3.1415926535897932384626433832795028841971693993751058209749445923078164062862089986280348253421170679; //approx of PI
    double order=1;
    double i, ERB;
    unsigned int j;
    double cf;  // temp. center freqeuenc
    double A11,A12,A13,A14, gain;
    double temp3, temp4, temp5, temp6, temp7, temp8;
    double T, B;
    complex<double> pole, tempComplex, theta, pole2,  temp1, temp2;
    zros.resize(4);
 for (i=0;i<=(nfft/2)+1;i++)
    {
        tempComplex=cDouble (tempComplex.real(), 2*pi*(i/nfft));
        ucirc.push_back(exp(tempComplex));
    }
    for (i=0;i<freqcenters.size();i++)
    {
        cf  =freqcenters[i];
        ERB=width*pow((pow((cf/EarQ),order) + pow(minBW,order)),(1/order));
        B = 2*pi*ERB*1.019;
        theta = cDouble (theta.real(),2*pi*cf/samplerate);
        pole=exp((-1*B)/samplerate)*exp(theta);
        pole2=pole;
        pole2=cDouble (pole2.real(),pole2.imag()*-1);
        // poles and zeros, following Malcolm's MakeERBFilter
        T=1/samplerate;
        temp1=cDouble (temp1.real(),4*cf*pi*T);
        temp2=cDouble (((-1)*B*T),2*cf*pi*T);
        temp3=pow(2,1.5);
        temp4=2*cf*pi*T;
        temp7=B*T;
        temp8=temp7*2.0;
        temp5=T*sin(temp4)/exp(temp7);
        temp6=2*T*cos(temp4)/exp(temp7);
        A11 = (temp6 + 2*sqrt(3+pow(2,1.5))*temp5)*0.5;
        A12 = (temp6 - 2*sqrt(3+pow(2,1.5))*temp5)*0.5;
        A13 = (temp6 + 2*sqrt(3-pow(2,1.5))*temp5)*0.5;
        A14 = (temp6 - 2*sqrt(3-pow(2,1.5))*temp5)*0.5;
        zros[0]=A11/T;
        zros[1]=A12/T;
        zros[2]=A13/T;
        zros[3]=A14/T;
        tempComplex=(-2.0*exp(temp1)*T +2.0*exp(temp2)*T*(cos(temp4) - sqrt(3-temp3)*sin(temp4)))*(-2.0*exp(temp1)*T + 2.0*exp(temp2)*T*(cos(temp4)+ sqrt(3-temp3) *sin(temp4)))*(-2.0*exp(temp1)*T + 2.0*exp(temp2)*T*(cos(temp4) -sqrt(3+temp3)*sin(temp4))) * (-2.0*exp(temp1)*T + 2.0*exp(temp2)*T* (cos(temp4) + sqrt(3+temp3)*sin(temp4)));
        gain=abs(tempComplex/pow((-2.0 / exp(temp8) - 2.0*exp(temp1) +2.0*(1.0 + exp(temp1))/exp(temp7)),4));

        for (j=0;j<wts.getSpalten();j++)
        {
            wts[i][j]=(pow(T,4)/gain)*abs(ucirc[j]-zros[0])*abs(ucirc[j]-zros[1])*abs(ucirc[j]-zros[2])*abs(ucirc[j]-zros[3])*pow(abs((pole-ucirc[j])*(pole2-ucirc[j])),-4);

        }
    }
    zros.clear();
    ucirc.clear();
    return wts;

}

/**********************************************************************************************
 *                                               FFT                                          *
 *                                                                                            *
 * Code by Sergey Chernenko                                                                   *
 * see "Fast Fourier transform  FFT" (http://www.librow.com/articles/article-10)             *
 *                                                                                            *
 * slight modification by Peter-Christian Quint for use with & vector<vector<complex>>        *
 *
 **********************************************************************************************/


/***********************************************************
 *                           fft                           *
 * preparation of input for calculating fft                *
 * calculate fft with fft-methode                          *
 * postprocessing data                                     *
 *                                                         *
 * INPUT:     Array2d<double> dFrames                      *
 * Output:    Array2d<double> newDframe                    *
 *                                                         *
 * size(newDframe)!=size(dFrames)  !                       *
 *                                                         *
 * June 2010  Peter-Christian Quint, pcq@gmx.de            *
 ***********************************************************/
Array2d<double>  GammatoneFB::doFFT(Array2d<double> dFrames)
{
    unsigned int i;
    unsigned int j, k;
    long int nextpower2=nextPower2(dFrames.getSpalten());
    cVec cFrame;
    Array2d<double> newDframe(dFrames.getZeilen(),nextpower2);
    cFrame.resize(nextpower2*2,0);                        // we make double size fft....

    for (i=0; i<dFrames.getZeilen();i++)
    {

        for (k=0;k<dFrames.getSpalten();k++)
        {
            cFrame[k]= dFrames[i][k];
        }
        for (k=dFrames.getSpalten();k<cFrame.size();k++)
        {
            cFrame[k]= 0;
        }
        FFT(cFrame,cFrame.size());                      //       FFT

        for (j=0;j<newDframe.getSpalten();j++)            // ...but we use only the half (fft is symmetric)
        {
            newDframe[i][j]=abs(cFrame[j]);
        }

    }

    cFrame.clear();


    return newDframe;
}
/******************************************************
 *                   nextpower2                       *
 * INPUT: (int)                                       *
 * OUTPUT: (int) the next power of two of input value *
 ******************************************************/
long int GammatoneFB::nextPower2(int x)
{
    long int n=2;
    while (n<x)
        n*=2;
    return n;
}


/**********************************************************************************************
 *                                               FFT                                          *
 *                                                                                            *
 * Code by Sergey Chernenko                                                                   *
 * see "Fast Fourier transform — FFT" (http://www.librow.com/articles/article-10)             *
 *                                                                                            *
 * slight modification by Peter-Christian Quint for use with & vector<vector<complex>>        *
 *											                                                  *
 **********************************************************************************************/

bool GammatoneFB::FFT(cVec &Data, const unsigned int N)
{
    //   Check input parameters
    if (Data.empty() || N < 1 || N & (N - 1))
        return false;
    //   Rearrange
    Rearrange(Data, N);
    //   Call FFT implementation
    Perform(Data, N);
    //   Succeeded
    return true;
}

//   INVERSE FOURIER TRANSFORM
//     Input  - input data
//     Output - transform result
//     N      - length of both input data and result
//     Scale  - if to scale result

//   Rearrange function

//   Inplace version of rearrange function
void GammatoneFB::Rearrange(cVec &Data, const unsigned int N)
{
    //   Swap position
    unsigned int Target = 0;
    //   Process all positions of input signal
    for (unsigned int Position = 0; Position < N; ++Position)
    {
        //   Only for not yet swapped entries
        if (Target > Position)
        {
            //   Swap entries
            cDouble Temp(Data[Target]);
            Data[Target] = Data[Position];
            Data[Position] = Temp;
        }
        //   Bit mask
        unsigned int Mask = N;
        //   While bit is set
        while (Target & (Mask >>= 1))
            //   Drop bit
            Target &= ~Mask;
        //   The current bit is 0 - set it
        Target |= Mask;
    }
}

//   FFT implementation
void GammatoneFB::Perform(cVec &Data, const unsigned int N, const bool Inverse /* = false */)
{
	const double pi = Inverse ? 3.14159265358979323846 : -3.14159265358979323846;
	//   Iteration through dyads, quadruples, octads and so on...
	for (unsigned int Step = 1; Step < N; Step <<= 1)
	{
		//   Jump to the next entry of the same transform factor
		const unsigned int Jump = Step << 1;
		//   Angle increment
		const double delta = pi / double(Step);
		//   Auxiliary sin(delta / 2)
		const double Sine = sin(delta * .5);
		//   Multiplier for trigonometric recurrence
		const cDouble Multiplier(-2. * Sine * Sine, sin(delta));
		//   Start value for transform factor, fi = 0
		cDouble Factor(1.);
		//   Iteration through groups of different transform factor
		for (unsigned int Group = 0; Group < Step; ++Group)
		{
			//   Iteration within group
			for (unsigned int Pair = Group; Pair < N; Pair += Jump)
			{
				//   Match position
				const unsigned int Match = Pair + Step;
				//   Second term of two-point transform
				const cDouble Product(Factor * Data[Match]);
				//   Transform for fi + pi
				Data[Match] = Data[Pair] - Product;
				//   Transform for fi
				Data[Pair] += Product;
			}
			//   Successive transform factor via trigonometric recurrence
			Factor = Multiplier * Factor + Factor;
		}
	}
}



